﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views;
using HeuristicLab.EvolutionTracking;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  /// <summary>
  /// This class implements the bottom-up tree distance described in the following paper: 
  /// G. Valiente, "An Efficient Bottom-up Distance Between Trees", http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.102.7958
  /// </summary>
  [StorableClass]
  [Item("Bottom-up distance calculator", "An operator that calculates the bottom-up distance between two symbolic expression trees.")]
  public class BottomUpDistanceCalculator : Item, IDistanceCalculator {
    public BottomUpDistanceCalculator() { }

    protected BottomUpDistanceCalculator(BottomUpDistanceCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BottomUpDistanceCalculator(this, cloner);
    }

    /// <summary>
    /// Creates a compact representation of the two trees as a directed acyclic graph
    /// </summary>
    /// <param name="t1">The first tree</param>
    /// <param name="t2">The second tree</param>
    /// <returns>The compacted DAG representing the two trees</returns>
    private static Dictionary<ISymbolicExpressionTreeNode, IVertex> Compact(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      var G = new DirectedGraph();
      var K = new Dictionary<ISymbolicExpressionTreeNode, IVertex>();
      var F = new DisjointUnion(t1, t2);
      var L = new Dictionary<string, IVertex>();
      var Children = new Dictionary<ISymbolicExpressionTreeNode, int>();

      var nodes = F.Nodes.ToList();

      // for all leaf labels l in F
      foreach (var l in nodes.Where(x => x.SubtreeCount == 0)) {
        var label = Label(l);
        var z = new Vertex { Content = l, Label = label };
        L[z.Label] = z;
        G.AddVertex(z);
      }

      var Q = new Queue<ISymbolicExpressionTreeNode>();

      // for all nodes
      foreach (var v in nodes) {
        Children[v] = v.SubtreeCount;
        if (v.SubtreeCount == 0) {
          Q.Enqueue(v);
        }
      }

      while (Q.Any()) {
        var v = Q.Dequeue();
        if (v.SubtreeCount == 0) {
          K[v] = L[Label(v)]; // 18
        } else {
          bool found = false;
          // for all nodes w in G in reverse order
          var vertices = G.Vertices.Reverse();
          foreach (var w in vertices) {
            if (Height(v) != Height(w) || v.SubtreeCount != w.OutDegree || Label(v) != w.Label)
              break;

            // sort V and W because we are dealing with unordered trees
            var V = v.Subtrees.Select(s => K[s]).OrderBy(x => x.Label);
            var W = w.OutArcs.Select(x => x.Target).OrderBy(x => x.Label);
            if (V.SequenceEqual(W)) {
              K[v] = w;
              found = true;
              break;
            }
          } // 32: end for

          if (!found) {
            var w = new Vertex { Content = v, Label = Label(v) };
            G.AddVertex(w);
            K[v] = w;

            foreach (var u in v.Subtrees) {
              G.AddArc(w, K[u]);
            } // 40: end for
          } // 41: end if
        } // 42: end if

        if (v.Parent != null) {
          var p = v.Parent;
          Children[p]--;
          if (Children[p] == 0)
            Q.Enqueue(p);
        }
      };

      return K;
    }

    private static Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode> Mapping(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2, Dictionary<ISymbolicExpressionTreeNode, IVertex> K) {
      var M = new Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode>(); // nodes of t1 => nodes of t2
      var M_ = new Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode>(); // nodes of t2 => nodes of t1
      var plist2 = t2.IterateNodesPrefix().ToList();
      foreach (var n in t1.IterateNodesBreadth()) {
        ISymbolicExpressionTreeNode v = n;
        if (!M.ContainsKey(v)) {
          var w = plist2.Last();
          var pw = plist2.IndexOf(w); // preorder index of node w
          foreach (var u in plist2.Where(x => K[x] == K[v])) {
            if (!M_.ContainsKey(u)) {
              w = u;
              // commented out test below since we are dealing with unordered trees
              //              if (plist2.IndexOf(u) < pw) {
              //                w = u;
              //              }
            }
          }
          if (K[v] == K[w]) {
            // simultaneous preorder traversal of the two subtrees
            var nodesV = v.IterateNodesPrefix().ToList();
            var nodesW = w.IterateNodesPrefix().ToList();
            for (int i = 0; i < Math.Min(nodesV.Count, nodesW.Count); ++i) {
              var s = nodesV[i];
              var t = nodesW[i];
              M[s] = t;
              M_[t] = s;
            }
          }
        }
      }
      return M;
    }

    // return a normalized distance measure in the interval (0,1)
    public double CalculateDistance(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      return BottomUpDistance(t1, t2) / (t1.Length + t2.Length);
    }

    public static double BottomUpDistance(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2, double p = 1, double q = 1, double r = 1) {
      var K = Compact(t1, t2);
      var M = Mapping(t1, t2, K);

      var keys = M.Keys.ToList();
      for (int j = 0; j < keys.Count; ++j) {
        var key = keys[j];
        var value = M[key];
        if (key.Parent != null && value.Parent != null && !M.ContainsKey(key.Parent) && Label(key.Parent).Equals(Label(value.Parent))) {
          M.Add(key.Parent, value.Parent);
          keys.Add(key.Parent);
        }
      }

      int d = t1.Length - M.Count;
      int s = M.Count(x => !Label(x.Key).Equals(Label(x.Value)));
      int i = t2.Length - M.Count;

      double distance = s * p + i * q + d * r;
      return distance;
    }

    private static string Label(ISymbolicExpressionTreeNode n) {
      return n.ToString();
    }

    private static int Height(IVertex v) {
      return Height((ISymbolicExpressionTreeNode)v.Content);
    }

    private static int Height(ISymbolicExpressionTreeNode n) {
      var p = n;
      while (p.Parent != null)
        p = p.Parent;
      return p.GetBranchLevel(n);
    }

    private class DisjointUnion : Tuple<ISymbolicExpressionTree, ISymbolicExpressionTree> {
      public DisjointUnion(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2)
        : base(t1, t2) {
      }

      public IEnumerable<ISymbolicExpressionTreeNode> Nodes {
        get {
          var r1 = Item1.Root;
          var r2 = Item2.Root;

          var nodes = r1.IterateNodesPostfix().Select(x => new { Node = x, Height = r1.GetBranchLevel(x) })
            .Concat(r2.IterateNodesPostfix().Select(x => new { Node = x, Height = r2.GetBranchLevel(x) }));

          return nodes.OrderByDescending(x => x.Height).Select(x => x.Node);

          //
          //          var p1 = r1.Parent;
          //          var p2 = r2.Parent;
          //
          //          var virtualRootSymbol = new StartSymbol();
          //          var virtualRootNode = virtualRootSymbol.CreateTreeNode();
          //
          //          virtualRootNode.AddSubtree(r1);
          //          virtualRootNode.AddSubtree(r2);
          //
          //          var nodes = virtualRootNode.IterateNodesPostfix().ToList();
          //
          //          virtualRootNode.RemoveSubtree(0);
          //          virtualRootNode.RemoveSubtree(0);
          //
          //          r1.Parent = p1;
          //          r2.Parent = p2;
          //
          //          for (int i = 0; i < nodes.Count - 1; ++i) { yield return nodes[i]; }
        }
      }
    }

    public static string RemoveFirstLines(string text, int linesCount) {
      var lines = Regex.Split(text, "\r\n|\r|\n").Skip(linesCount);
      return string.Join(Environment.NewLine, lines.ToArray());
    }

    // draw the mapping between t1 and t2 as a tikz picture
    private static string FormatMapping(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2, Dictionary<ISymbolicExpressionTreeNode, ISymbolicExpressionTreeNode> map) {
      var formatter = new SymbolicExpressionTreeLatexFormatter();
      var sb = new StringBuilder();

      string s1, s2;
      var m1 = formatter.Format(t1, out s1);
      var m2 = formatter.Format(t2, out s2, 20);

      // skip first 6 lines of s2
      s2 = RemoveFirstLines(s2, 6);

      sb.Append(s1);
      sb.Append(s2);

      foreach (var p in map) {
        var id1 = m1[p.Key];
        var id2 = m2[p.Value];

        sb.Append(string.Format(CultureInfo.InvariantCulture, "\\path[draw,->] ({0}) edge[bend left,dashed] ({1});" + Environment.NewLine, id1, id2));
      }

      return sb.ToString();
    }

  }
}
